# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import logging
from functools import partial

from botocore import xform_name

from .action import ServiceAction
from .action import WaiterAction
from .base import ResourceMeta, ServiceResource
from .collection import CollectionFactory
from .model import ResourceModel
from .response import build_identifiers, ResourceHandler
from ..exceptions import ResourceLoadException


logger = logging.getLogger(__name__)


class ResourceFactory(object):
    """
    A factory to create new :py:class:`~boto3.resources.base.ServiceResource`
    classes from a :py:class:`~boto3.resources.model.ResourceModel`. There are
    two types of lookups that can be done: one on the service itself (e.g. an
    SQS resource) and another on models contained within the service (e.g. an
    SQS Queue resource).
    """
    def __init__(self):
        self._collection_factory = CollectionFactory()

    def load_from_definition(self, service_name, resource_name, model,
                             resource_defs, service_model):
        """
        Loads a resource from a model, creating a new
        :py:class:`~boto3.resources.base.ServiceResource` subclass
        with the correct properties and methods, named based on the service
        and resource name, e.g. EC2.Instance.

        :type service_name: string
        :param service_name: Name of the service to look up
        :type resource_name: string
        :param resource_name: Name of the resource to look up. For services,
                              this should match the ``service_name``.
        :type model: dict
        :param model: The service or resource definition.
        :type resource_defs: dict
        :param resource_defs: The service's resource definitions, used to load
                              subresources (e.g. ``sqs.Queue``).
        :type service_model: ``botocore.model.ServiceModel``
        :param service_model: The Botocore service model, required only if the
                              resource shape contains members. This is used to
                              expose lazy-loaded attributes on the resource.
        :rtype: Subclass of :py:class:`~boto3.resources.base.ServiceResource`
        :return: The service or resource class.
        """
        # Set some basic info
        meta = ResourceMeta(service_name)
        attrs = {
            'meta': meta,
        }

        logger.debug('Loading %s:%s', service_name, resource_name)

        resource_model = ResourceModel(resource_name, model, resource_defs)

        self._load_identifiers(attrs, meta, resource_model)
        self._load_actions(attrs, resource_model, resource_defs,
                           service_model)
        self._load_attributes(attrs, meta, resource_model, service_model)
        self._load_collections(attrs, resource_model, resource_defs,
                               service_model)
        self._load_has_relations(attrs, service_name, resource_name,
                                 resource_model, resource_defs, service_model)
        self._load_waiters(attrs, resource_model)

        # Create the name based on the requested service and resource
        cls_name = resource_name
        if service_name != resource_name:
            cls_name = service_name + '.' + cls_name

        return type(str(cls_name), (ServiceResource,), attrs)

    def _load_identifiers(self, attrs, meta, model):
        """
        Populate required identifiers. These are arguments without which
        the resource cannot be used. Identifiers become arguments for
        operations on the resource.
        """
        for identifier in model.identifiers:
            snake_cased = xform_name(identifier.name)
            snake_cased = self._check_allowed_name(
                attrs, snake_cased, 'identifier', model.name)
            meta.identifiers.append(snake_cased)
            attrs[snake_cased] = None

    def _load_actions(self, attrs, model, resource_defs, service_model):
        """
        Actions on the resource become methods, with the ``load`` method
        being a special case which sets internal data for attributes, and
        ``reload`` is an alias for ``load``.
        """
        if model.load:
            attrs['load'] = self._create_action(
                'load', model.load, resource_defs, service_model,
                is_load=True)
            attrs['reload'] = attrs['load']

        for action in model.actions:
            snake_cased = xform_name(action.name)
            snake_cased = self._check_allowed_name(
                attrs, snake_cased, 'action', model.name)
            attrs[snake_cased] = self._create_action(snake_cased,
                action, resource_defs, service_model)

    def _load_attributes(self, attrs, meta, model, service_model):
        """
        Load resource attributes based on the resource shape. The shape
        name is referenced in the resource JSON, but the shape itself
        is defined in the Botocore service JSON, hence the need for
        access to the ``service_model``.
        """
        if model.shape:
            shape = service_model.shape_for(model.shape)

            for name, member in shape.members.items():
                snake_cased = xform_name(name)
                if snake_cased in meta.identifiers:
                    # Skip identifiers, these are set through other means
                    continue

                snake_cased = self._check_allowed_name(
                    attrs, snake_cased, 'attribute', model.name)
                attrs[snake_cased] = self._create_autoload_property(name,
                    snake_cased)

    def _load_collections(self, attrs, model, resource_defs, service_model):
        """
        Load resource collections from the model. Each collection becomes
        a :py:class:`~boto3.resources.collection.CollectionManager` instance
        on the resource instance, which allows you to iterate and filter
        through the collection's items.
        """
        for collection_model in model.collections:
            snake_cased = xform_name(collection_model.name)
            snake_cased = self._check_allowed_name(
                attrs, snake_cased, 'collection', model.name)

            attrs[snake_cased] = self._create_collection(
                attrs['meta'].service_name, model.name, snake_cased,
                collection_model, resource_defs, service_model)

    def _load_has_relations(self, attrs, service_name, resource_name,
                            model, resource_defs, service_model):
        """
        Load related resources, which are defined via a ``has``
        relationship but conceptually come in two forms:

        1. A reference, which is a related resource instance and can be
           ``None``, such as an EC2 instance's ``vpc``.
        2. A subresource, which is a resource constructor that will always
           return a resource instance which shares identifiers/data with
           this resource, such as ``s3.Bucket('name').Object('key')``.
        """
        for reference in model.references:
            # This is a dangling reference, i.e. we have all
            # the data we need to create the resource, so
            # this instance becomes an attribute on the class.
            snake_cased = xform_name(reference.name)
            snake_cased = self._check_allowed_name(
                attrs, snake_cased, 'reference', model.name)
            attrs[snake_cased] = self._create_reference(
                reference.resource.type, snake_cased, reference,
                service_name, resource_name, model, resource_defs,
                service_model)

        for subresource in model.subresources:
            # This is a sub-resource class you can create
            # by passing in an identifier, e.g. s3.Bucket(name).
            name = subresource.resource.type
            attrs[subresource.name] = self._create_class_partial(
                name, subresource, service_name, resource_name, model,
                resource_defs, service_model)


    def _load_waiters(self, attrs, model):
        """
        Load resource waiters from the model. Each waiter allows you to
        wait until a resource reaches a specific state by polling the state
        of the resource.
        """
        for waiter in model.waiters:
            snake_cased = xform_name(waiter.resource_waiter_name)
            snake_cased = self._check_allowed_name(
                attrs, snake_cased, 'waiter', model.name)
            attrs[snake_cased] = self._create_waiter(waiter, snake_cased)

    def _check_allowed_name(self, attrs, name, category, resource_name):
        """
        Determine if a given name is allowed on the instance, and if not,
        then raise an exception. This prevents public attributes of the
        class from being clobbered, e.g. since we define ``Resource.meta``,
        no identifier may be named ``meta``. Another example: no action
        named ``queue_items`` may be added after an identifier of the same
        name has been added.

        One attempt is made in the event of a collision to remedy the
        situation. The ``category`` is appended to the name and the
        check is performed again. For example, if an action named
        ``get_frobs`` fails the test, then we try ``get_frobs_action``
        after logging a warning.

        :raises: ValueError
        """
        if name in attrs:
            logger.warning('%s `%s` would clobber existing %s'
                           ' resource attribute, going to try'
                           ' %s instead...', category, name,
                           resource_name, name + '_' + category)
            # TODO: Move this logic into the model and strictly
            #       define the loading order of categories. This
            #       will make documentation much simpler.
            name = name + '_' + category

        if name in attrs:
            raise ValueError('{0} `{1}` would clobber existing '
                             '{2} resource attribute'.format(
                                category, name, resource_name))

        return name

    def _create_autoload_property(factory_self, name, snake_cased):
        """
        Creates a new property on the resource to lazy-load its value
        via the resource's ``load`` method (if it exists).
        """
        # The property loader will check to see if this resource has already
        # been loaded and return the cached value if possible. If not, then
        # it first checks to see if it CAN be loaded (raise if not), then
        # calls the load before returning the value.
        def property_loader(self):
            if self.meta.data is None:
                if hasattr(self, 'load'):
                    self.load()
                else:
                    raise ResourceLoadException(
                        '{0} has no load method'.format(self.__class__.__name__))

            return self.meta.data.get(name)

        property_loader.__name__ = str(snake_cased)
        property_loader.__doc__ = 'TODO'
        return property(property_loader)

    def _create_waiter(factory_self, waiter_model, snake_cased):
        """
        Creates a new wait method for each resource where both a waiter and
        resource model is defined.
        """
        waiter = WaiterAction(waiter_model, waiter_resource_name=snake_cased)
        def do_waiter(self, *args, **kwargs):
            waiter(self, *args, **kwargs)

        do_waiter.__name__ = str(snake_cased)
        do_waiter.__doc__ = 'TODO'
        return do_waiter

    def _create_collection(factory_self, service_name, resource_name,
                           snake_cased, collection_model,
                           resource_defs, service_model):
        """
        Creates a new property on the resource to lazy-load a collection.
        """
        cls = factory_self._collection_factory.load_from_definition(
            service_name, resource_name, collection_model.name,
            collection_model, resource_defs)

        def get_collection(self):
            return cls(collection_model, self, factory_self,
                       resource_defs, service_model)

        get_collection.__name__ = str(snake_cased)
        get_collection.__doc__ = 'TODO'
        return property(get_collection)

    def _create_reference(factory_self, name, snake_cased, reference,
                          service_name, resource_name, model, resource_defs,
                          service_model):
        """
        Creates a new property on the resource to lazy-load a reference.
        """
        # References are essentially an action with no request
        # or response, so we can re-use the response handlers to
        # build up resources from identifiers and data members.
        handler = ResourceHandler('', factory_self, resource_defs,
                                  service_model, reference.resource)

        def get_reference(self):
            # We need to lazy-evaluate the reference to handle circular
            # references between resources. We do this by loading the class
            # when first accessed.
            # First, though, we need to see if we have the required
            # identifiers to instantiate the resource reference.
            return handler(self, {}, {})

        get_reference.__name__ = str(snake_cased)
        get_reference.__doc__ = 'TODO'
        return property(get_reference)

    def _create_class_partial(factory_self, name, subresource,
                              service_name, resource_name, model,
                              resource_defs, service_model):
        """
        Creates a new method which acts as a functools.partial, passing
        along the instance's low-level `client` to the new resource
        class' constructor.
        """
        # We need a new method here because we want access to the
        # instance's client.
        def create_resource(self, *args, **kwargs):
            positional_args = []

            # We lazy-load the class to handle circular references.
            resource_cls = factory_self.load_from_definition(
                service_name, name, resource_defs.get(name, {}),
                resource_defs, service_model)

            # Assumes that identifiers are in order, which lets you do
            # e.g. ``sqs.Queue('foo').Message('bar')`` to create a new message
            # linked with the ``foo`` queue and which has a ``bar`` receipt
            # handle. If we did kwargs here then future positional arguments
            # would lead to failure.
            identifiers = subresource.resource.identifiers
            if identifiers is not None:
                for identifier, value in build_identifiers(identifiers, self):
                    positional_args.append(value)

            return partial(resource_cls, *positional_args,
                client=self.meta.client)(*args, **kwargs)

        create_resource.__name__ = str(name)
        create_resource.__doc__ = 'TODO'
        return create_resource

    def _create_action(factory_self, snake_cased, action_model, resource_defs,
                       service_model, is_load=False):
        """
        Creates a new method which makes a request to the underlying
        AWS service.
        """
        # Create the action in in this closure but before the ``do_action``
        # method below is invoked, which allows instances of the resource
        # to share the ServiceAction instance.
        action = ServiceAction(action_model, factory=factory_self,
            resource_defs=resource_defs, service_model=service_model)

        # A resource's ``load`` method is special because it sets
        # values on the resource instead of returning the response.
        if is_load:
            # We need a new method here because we want access to the
            # instance via ``self``.
            def do_action(self, *args, **kwargs):
                response = action(self, *args, **kwargs)
                self.meta.data = response
        else:
            # We need a new method here because we want access to the
            # instance via ``self``.
            def do_action(self, *args, **kwargs):
                response = action(self, *args, **kwargs)

                if hasattr(self, 'load'):
                    # Clear cached data. It will be reloaded the next
                    # time that an attribute is accessed.
                    # TODO: Make this configurable in the future?
                    self.meta.data = None

                return response

        do_action.__name__ = str(snake_cased)
        do_action.__doc__ = 'TODO'
        return do_action



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os

import botocore.session

import boto3

from .exceptions import NoVersionFound
from .resources.factory import ResourceFactory


class Session(object):
    """
    A session stores configuration state and allows you to create service
    clients and resources.

    :type aws_access_key_id: string
    :param aws_access_key_id: AWS access key ID
    :type aws_secret_access_key: string
    :param aws_secret_access_key: AWS secret access key
    :type aws_session_token: string
    :param aws_session_token: AWS temporary session token
    :type region_name: string
    :param region_name: Default region when creating new connections
    :type botocore_session: botocore.session.Session
    :param botocore_session: Use this Botocore session instead of creating
                             a new default one.
    """
    def __init__(self, aws_access_key_id=None, aws_secret_access_key=None,
                 aws_session_token=None, region_name=None,
                 botocore_session=None):
        if botocore_session is not None:
            self._session = botocore_session
        else:
            # Create a new default session
            self._session = botocore.session.Session()

        # Setup custom user-agent string if it isn't already customized
        if self._session.user_agent_name == 'Botocore':
            botocore_info = 'Botocore/{0}'.format(
                self._session.user_agent_version)
            if self._session.user_agent_extra:
                self._session.user_agent_extra += ' ' + botocore_info
            else:
                self._session.user_agent_extra = botocore_info
            self._session.user_agent_name = 'Boto3'
            self._session.user_agent_version = boto3.__version__

        if aws_access_key_id or aws_secret_access_key or aws_session_token:
            self._session.set_credentials(aws_access_key_id,
                aws_secret_access_key, aws_session_token)

        if region_name is not None:
            self._session.set_config_variable('region', region_name)

        self.resource_factory = ResourceFactory()
        self._setup_loader()

    def __repr__(self):
        return 'Session(region={0})'.format(
            repr(self._session.get_config_variable('region')))

    def _setup_loader(self):
        """
        Setup loader paths so that we can load resources.
        """
        self._loader = self._session.get_component('data_loader')
        self._loader.data_path = ':'.join(
            [self._loader.data_path,
             os.path.join(os.path.dirname(__file__), 'data',
                          'resources')]).strip(':')

    def _get_resource_files(self):
        """
        This generator yields paths to resource files in the loader's
        search paths. Specifically, it looks for files that end with
        ``.resources.json`` in any of the search paths, but does not
        recursively search the paths.
        """
        for path in self._loader.get_search_paths():
            if not os.path.isdir(path) or not os.path.exists(path):
                continue

            items = os.listdir(path)
            for entry in [i for i in items if i.endswith('.resources.json')]:
                yield entry

    def _find_latest_version(self, service_name):
        """
        Find the latest resource version of a given service if it exists,
        otherwise raises an exception.

        TODO: Merge this logic upstream into Botocore if possible. Botocore
              depends on a different directory layout at the moment.

        :rtype: string
        :return: Version string like 'YYYY-MM-DD'
        :raises: NoVersionFound
        """
        filtered = []
        for path in self._get_resource_files():
            if path.startswith(service_name + '-'):
                filtered.append(path)

        try:
            # ['s3-2006-03-01.resources.json', ...] => '2006-03-01'
            # Hard coded offsets below pull out just the date string
            start = len(service_name)
            return max([i[start + 1:start + 11] for i in filtered])
        except ValueError:
            raise NoVersionFound(service_name)

    def get_available_services(self):
        """
        Get a list of available services that can be loaded as low-level
        clients via :py:meth:`Session.client`.

        :rtype: list
        :return: List of service names
        """
        return self._session.get_available_services()

    def get_available_resources(self):
        """
        Get a list of available services that can be loaded as resource
        clients via :py:meth:`Session.resource`.

        :rtype: list
        :return: List of service names
        """
        service_names = set()

        for path in self._get_resource_files():
            # 'foo-bar-2006-03-01' => 'foo-bar'
            service_names.add('-'.join(path.split('-')[:-3]))

        return list(service_names)

    def client(self, service_name, region_name=None, api_version=None,
               use_ssl=True, verify=None, endpoint_url=None,
               aws_access_key_id=None, aws_secret_access_key=None,
               aws_session_token=None):
        """
        Create a low-level service client by name.

        :type service_name: string
        :param service_name: The name of a service, e.g. 's3' or 'ec2'. You
            can get a list of available services via
            :py:meth:`get_available_services`.

        :type region_name: string
        :param region_name: The name of the region associated with the client.
            A client is associated with a single region.

        :type api_version: string
        :param api_version: The API version to use.  By default, botocore will
            use the latest API version when creating a client.  You only need
            to specify this parameter if you want to use a previous API version
            of the client.

        :type use_ssl: boolean
        :param use_ssl: Whether or not to use SSL.  By default, SSL is used.  Note that
            not all services support non-ssl connections.

        :type verify: boolean/string
        :param verify: Whether or not to verify SSL certificates.  By default SSL certificates
            are verified.  You can provide the following values:

            * False - do not validate SSL certificates.  SSL will still be
              used (unless use_ssl is False), but SSL certificates
              will not be verified.
            * path/to/cert/bundle.pem - A filename of the CA cert bundle to
              uses.  You can specify this argument if you want to use a different
              CA cert bundle than the one used by botocore.

        :type endpoint_url: string
        :param endpoint_url: The complete URL to use for the constructed client.
            Normally, botocore will automatically construct the appropriate URL
            to use when communicating with a service.  You can specify a
            complete URL (including the "http/https" scheme) to override this
            behavior.  If this value is provided, then ``use_ssl`` is ignored.

        :type aws_access_key_id: string
        :param aws_access_key_id: The access key to use when creating
            the client.  This is entirely optional, and if not provided,
            the credentials configured for the session will automatically
            be used.  You only need to provide this argument if you want
            to override the credentials used for this specific client.

        :type aws_secret_access_key: string
        :param aws_secret_access_key: The secret key to use when creating
            the client.  Same semantics as aws_access_key_id above.

        :type aws_session_token: string
        :param aws_session_token: The session token to use when creating
            the client.  Same semantics as aws_access_key_id above.

        :return: Service client instance
        """
        return self._session.create_client(
            service_name, region_name=region_name, api_version=api_version,
            use_ssl=use_ssl, verify=verify, endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token)

    def resource(self, service_name, region_name=None, api_version=None,
               use_ssl=True, verify=None, endpoint_url=None,
               aws_access_key_id=None, aws_secret_access_key=None,
               aws_session_token=None):
        """
        Create a resource service client by name.

        :type service_name: string
        :param service_name: The name of a service, e.g. 's3' or 'ec2'. You
            can get a list of available services via
            :py:meth:`get_available_resources`.

        :type region_name: string
        :param region_name: The name of the region associated with the client.
            A client is associated with a single region.

        :type api_version: string
        :param api_version: The API version to use.  By default, botocore will
            use the latest API version when creating a client.  You only need
            to specify this parameter if you want to use a previous API version
            of the client.

        :type use_ssl: boolean
        :param use_ssl: Whether or not to use SSL.  By default, SSL is used.  Note that
            not all services support non-ssl connections.

        :type verify: boolean/string
        :param verify: Whether or not to verify SSL certificates.  By default SSL certificates
            are verified.  You can provide the following values:

            * False - do not validate SSL certificates.  SSL will still be
              used (unless use_ssl is False), but SSL certificates
              will not be verified.
            * path/to/cert/bundle.pem - A filename of the CA cert bundle to
              uses.  You can specify this argument if you want to use a different
              CA cert bundle than the one used by botocore.

        :type endpoint_url: string
        :param endpoint_url: The complete URL to use for the constructed client.
            Normally, botocore will automatically construct the appropriate URL
            to use when communicating with a service.  You can specify a
            complete URL (including the "http/https" scheme) to override this
            behavior.  If this value is provided, then ``use_ssl`` is ignored.

        :type aws_access_key_id: string
        :param aws_access_key_id: The access key to use when creating
            the client.  This is entirely optional, and if not provided,
            the credentials configured for the session will automatically
            be used.  You only need to provide this argument if you want
            to override the credentials used for this specific client.

        :type aws_secret_access_key: string
        :param aws_secret_access_key: The secret key to use when creating
            the client.  Same semantics as aws_access_key_id above.

        :type aws_session_token: string
        :param aws_session_token: The session token to use when creating
            the client.  Same semantics as aws_access_key_id above.

        :return: Subclass of :py:class:`~boto3.resources.base.ServiceResource`
        """
        # Creating a new resource instance requires the low-level client
        # and service model, the resource version and resource JSON data.
        # We pass these to the factory and get back a class, which is
        # instantiated on top of the low-level client.
        client = self.client(
            service_name, region_name=region_name, api_version=api_version,
            use_ssl=use_ssl, verify=verify, endpoint_url=endpoint_url,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token)
        service_model = self._session.get_service_model(service_name)
        version = self._find_latest_version(service_name)
        model = self._loader.load_data(
            '{0}-{1}.resources'.format(service_name, version))
        cls = self.resource_factory.load_from_definition(
            service_name, service_name, model['service'], model['resources'],
            service_model)
        return cls(client=client)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

"""
This module is used to generate both high and low-level reference
documentation for services. Currently, it does this by inspecting
the service and resource models, as well as instantiating dummy
clients to introspect some values. It is likely to change
significantly in the future!

Currently this is not used for docstrings, just for the Sphinx
documentation. RST is generated which Sphinx turns into HTML.

The generated output can be found here:

    http://boto3.readthedocs.org/en/latest/

"""

import json
import os

import botocore.session

from botocore import xform_name
from bcdoc.restdoc import ReSTDocument

import boto3

from .resources.model import ResourceModel


def py_type_name(type_name):
    """
    Get the Python type name for a given model type.

        >>> py_type_name('list')
        'list'
        >>> py_type_name('structure')
        'dict'

    :rtype: string
    """
    return {
        'character': 'string',
        'double': 'float',
        'long': 'integer',
        'map': 'dict',
        'structure': 'dict',
        'timestamp': 'datetime',
    }.get(type_name, type_name)


def py_default(type_name):
    """
    Get the Python default value for a given model type, useful
    for generated examples.

        >>> py_default('string')
        '\'string\''
        >>> py_default('list')
        '[...]'
        >>> py_default('unknown')
        '...'

    :rtype: string
    """
    return {
        'double': '123.0',
        'long': '123',
        'integer': '123',
        'string': "'string'",
        'blob': "b'bytes'",
        'list': '[...]',
        'map': '{...}',
        'structure': '{...}',
        'timestamp': 'datetime(2015, 1, 1)',
    }.get(type_name, '...')


def html_to_rst(html, indent=0, indentFirst=False):
    """
    Use bcdoc to convert html to rst.

    :type html: string
    :param html: Input HTML to be converted
    :type indent: int
    :param indent: Number of spaces to indent each line
    :type indentFirst: boolean
    :param indentFirst: Whether to indent the first line
    :rtype: string
    """
    doc = ReSTDocument()

    # TODO: Remove me, temp workaround to fix doc building
    # because of smart quotes that aren't currently supported.
    html = html.replace(u'\u2019', "'")
    html = html.replace(u'\u2014', '-')

    doc.include_doc_string(html)
    rst = doc.getvalue().decode('utf-8')

    if indent:
        rst = '\n'.join([(' ' * indent) + line for line in rst.splitlines()])

        if not indentFirst:
            rst = rst.strip()

    return rst


def docs_for(service_name):
    """
    Generate reference documentation (low and high level) for a service
    by name. This generates docs for the latest available version.

    :type service_name: string
    :param service_name: The service short-name, like 'ec2'
    :rtype: string
    """
    session = botocore.session.get_session()
    service_model = session.get_service_model(service_name)

    print('Processing {0}-{1}'.format(service_name, service_model.api_version))

    # The following creates an official name of 'Amazon Simple Storage
    # Service (S3)' our of 'Amazon Simple Storage Service' and 'Amazon S3'.
    # It tries to be smart, so for `Amazon DynamoDB' and 'DynamoDB' we would
    # get an official name of 'Amazon DynamoDB'.
    official_name = service_model.metadata.get('serviceFullName')
    short_name = service_model.metadata.get('serviceAbbreviation', '')
    if short_name.startswith('Amazon'):
        short_name = short_name[7:]
    if short_name.startswith('AWS'):
        short_name = short_name[4:]
    if short_name and short_name.lower() not in official_name.lower():
        official_name += ' ({0})'.format(short_name)

    docs = '{0}\n{1}\n\n'.format(official_name, '=' * len(official_name))

    docs += '.. contents:: Table of Contents\n   :depth: 2\n\n'

    docs += document_client(service_name, official_name, service_model)
    docs += document_client_waiter(session, official_name, service_name,
                                   service_model)

    filename = (os.path.dirname(__file__) + '/data/resources/'
                '{0}-{1}.resources.json').format(service_name,
                                                 service_model.api_version)
    # We can't use a set here because dicts aren't hashable!
    models = {}
    if os.path.exists(filename):
        data = json.load(open(filename))
        model = ResourceModel(service_name, data['service'], data['resources'])

        for collection_model in model.collections:
            collection_model.parent_name = model.name
            models[collection_model.name] = {
                'type': 'collection',
                'model': collection_model
            }

        docs += document_resource(service_name, official_name, model,
                                  service_model, session)

        # First, collect all the models...
        for name, model in sorted(data['resources'].items(),
                                  key=lambda i:i[0]):
            resource_model = ResourceModel(name, model, data['resources'])
            if name not in models:
                models[name] = {'type': 'resource', 'model': resource_model}

            for collection_model in resource_model.collections:
                collection_model.parent_name = xform_name(resource_model.name)

                cname = collection_model.name + 'CollectionManager'
                if cname not in models:
                    models[cname] = {'type': 'collection',
                                     'model': collection_model}

        # Then render them out in alphabetical order.
        for name, item in sorted(models.items()):
            model = item['model']
            if item['type'] == 'resource':
                docs += document_resource(service_name, official_name,
                                          model, service_model, session)
            elif item['type'] == 'collection':
                docs += document_collection(
                    service_name, official_name, model,
                    model.resource.model, service_model)

    return docs

def document_client(service_name, official_name, service_model):
    """
    Generate low-level client documentation for a service. This generates
    documentation for all available operations.
    """
    docs = 'Client\n------\n\n'
    docs += '.. py:class:: {0}.Client\n\n'.format(service_name)
    docs += '   A low-level client representing {0}::\n\n'.format(
        official_name)
    docs += '       import boto3\n\n'
    docs += '       {service} = boto3.client(\'{service}\')\n\n'.format(
        service=service_name)

    # TODO: Get this information from the model somehow in the future.
    #       For now creating and introspecing a client is a quick and
    #       dirty way to access waiters/paginators.
    client = boto3.client(service_name, aws_access_key_id='dummy',
                          aws_secret_access_key='dummy',
                          region_name='us-east-1')

    wdoc = ''
    if client.waiter_names:
        # This gets included in alphabetical order below!
        wdoc += '   .. py:method:: get_waiter(name)\n\n'
        wdoc += '      Get a waiter by name. Available waiters:\n\n'
        for waiter in client.waiter_names:
            wdoc += '      * `{0}`_\n'.format(waiter)
        wdoc += '\n'

    waiter_included = False
    for operation_name in service_model.operation_names:
        if not waiter_included and xform_name(operation_name) > 'get_waiter':
            docs += wdoc
            waiter_included = True

        operation = service_model.operation_model(operation_name)
        docs += document_operation(
            operation, service_name,
            paginated=client.can_paginate(xform_name(operation_name)),
            example_instance='client', example_response='response')

    return docs

def document_client_waiter(session, official_name, service_name,
                           service_model):
    client = boto3.client(service_name, aws_access_key_id='dummy',
                          aws_secret_access_key='dummy',
                          region_name='us-east-1')
    waiter_spec_doc = ''
    if client.waiter_names:
        waiter_spec_doc = 'Waiter\n------\n\n'
        service_waiter_model = session.get_waiter_model(service_name)
        for waiter in service_waiter_model.waiter_names:
            snake_cased = xform_name(waiter)
            waiter_spec_doc += '{0}\n{1}\n\n'.format(snake_cased,
                '~' * len(snake_cased))
            waiter_model = service_waiter_model.get_waiter(waiter)
            operation_model = service_model.operation_model(
                waiter_model.operation)
            description = (
                '      This polls :py:meth:`{0}.Client.{1}` every {2} '
                'seconds until a successful state is reached. An error is '
                'returned after {3} failed checks.'.format(
                    service_name, xform_name(waiter_model.operation),
                    waiter_model.delay, waiter_model.max_attempts)
            )
            waiter_spec_doc += document_operation(
                operation_model=operation_model, service_name=service_name,
                operation_name='wait', rtype=None, description=description,
                example_instance='client.get_waiter(\'{0}\')'.format(
                    snake_cased))
        waiter_spec_doc += '\n'

    return waiter_spec_doc

def document_resource(service_name, official_name, resource_model,
                      service_model, session):
    """
    Generate reference documentation from a resource model.
    """
    model_name = resource_model.name
    title = model_name
    is_service_resource = False
    if model_name == service_name:
        model_name = 'Service'
        title = 'Service Resource'
        is_service_resource = True
    docs = '{0}\n{1}\n\n'.format(title, '-' * len(title))
    docs += '.. py:class:: {0}.{1}({2})\n\n'.format(
        service_name, model_name, ', '.join(
            [xform_name(i.name) for i in resource_model.identifiers]))

    if is_service_resource:
        docs += ('   A resource representing {0}::\n\n').format(official_name)
        docs += '       import boto3\n\n'
        docs += '       {service} = boto3.resource(\'{service}\')\n\n'.format(
            service=service_name)
    else:
        identifiers = ', '.join(
            ["'{0}'".format(xform_name(i.name)) for i in
             resource_model.identifiers])
        docs += ('   A resource representing an {0} {1}::\n\n').format(
            official_name, model_name)
        docs += '       import boto3\n\n'
        docs += ('       {service} = boto3.resource(\'{service}\')\n'
                 '       {var} = {service}.{model}({identifiers})\n\n').format(
                    var=xform_name(model_name), service=service_name,
                    model=model_name, identifiers=identifiers)

    if not is_service_resource:
        docs += ('   .. rst-class:: admonition-title\n\n   Attributes &'
                 'Identifiers\n\n   Attributes & identifiers provide access'
                 ' to the properties of a resource. Attributes are lazy-'
                 'loaded the first time one is accessed via the'
                 ' :py:meth:`load` method, if it exists.\n\n'
                 '   Identifiers:\n\n')

        for identifier in sorted(resource_model.identifiers,
                                 key=lambda i:i.name):
            docs += ('   .. py:attribute:: {0}\n\n      (``string``,'
                     ' **identifier**) The {1}\'s {2} identifier. This'
                     ' attribute **must** be set for the actions below to'
                     ' work.\n\n'.format(
                        xform_name(identifier.name), resource_model.name,
                        identifier.name))

        docs += '\n\n'

        if resource_model.shape:
            docs += '   Attributes:\n\n'
            shape = service_model.shape_for(resource_model.shape)

            for name, member in sorted(shape.members.items()):
                docs += ('   .. py:attribute:: {0}\n\n      (``{1}``)'
                         ' {2}\n\n').format(
                    xform_name(name), py_type_name(member.type_name),
                    html_to_rst(member.documentation, indent=6))

    docs += ('   .. rst-class:: admonition-title\n\n   Actions\n\n   Actions'
             ' call operations on resources, automatically handling the'
             ' passing in of arguments set from identifiers and some'
             ' attributes.\n\n')
    for action in sorted(resource_model.actions, key=lambda i:i.name):
        docs += document_action(action, service_name, resource_model,
                                service_model)

    if resource_model.subresources:
        docs += ('   .. rst-class:: admonition-title\n\n   Sub-resources\n\n'
                 '   Sub-resources are methods that create a new instance of a'
                 ' child resource. This resource\'s identifiers get passed'
                 ' along to the child.\n\n')

        preset = len(resource_model.identifiers)

        if resource_model.subresources:
            for subresource in sorted(resource_model.subresources,
                                      key=lambda i: i.name):
                identifiers = [
                    xform_name(i.target) for i in \
                    subresource.resource.identifiers if i.source == 'input']
                docs += '   .. py:method:: {0}({1})\n\n'.format(
                    subresource.name,
                    ', '.join(identifiers))
                docs += ('      Create a :py:class:`{0}.{1}`'
                         ' instance.\n\n').format(service_name,
                                                  subresource.resource.type)

        docs += '\n\n'

    if resource_model.references:
        docs += ('   .. rst-class:: admonition-title\n\n   References\n\n'
                 '   References are related resource instances that have'
                 ' a belongs-to relationship.\n\n')
        for ref in sorted(resource_model.references, key=lambda i: i.name):
            docs += ('   .. py:attribute:: {0}\n\n      '
                     '(:py:class:`{1}.{2}`) The related {3} if set,'
                     ' otherwise ``None``.\n\n').format(
                        xform_name(ref.name), service_name,
                        ref.resource.type, ref.resource.type)

    if resource_model.collections:
        docs += ('   .. rst-class:: admonition-title\n\n   Collections\n\n'
                 '   Collections provide an interface to iterate and'
                 ' manipulate groups of resources.\n\n')
        for collection in sorted(resource_model.collections,
                                 key=lambda i: i.name):
            docs += ('   .. py:attribute:: {0}\n\n      '
                     '(:py:class:`{1}.{2}CollectionManager`)'
                     ' A collection of :py:class:`{3}.{4}` instances. This'
                     ' collection uses the :py:meth:`{5}.Client.{6}` operation'
                     ' to get items.\n\n').format(
                        xform_name(collection.name), service_name,
                        collection.name, service_name,
                        collection.resource.type, service_name,
                        xform_name(collection.request.operation))

    if resource_model.waiters:
        docs += ('   .. rst-class:: admonition-title\n\n   Waiters\n\n'
                 '   Waiters provide an interface to wait for a resource'
                 ' to reach a specific state.\n\n')
        service_waiter_model = session.get_waiter_model(service_name)
        for waiter in sorted(resource_model.waiters,
                             key=lambda i: i.resource_waiter_name):
            docs += document_waiter(waiter, service_name, resource_model,
                                    service_model, service_waiter_model)

    return docs

def document_collection(service_name, official_name, collection_model,
                        resource_model, service_model):
    """
    Generate reference documentation about a collection and any
    batch actions it might have.
    """
    title = collection_model.name + 'Collection'
    docs = '{0}\n{1}\n\n'.format(title, '-' * len(title))
    docs += '.. py:class:: {0}.{1}CollectionManager()\n\n'.format(
        service_name, collection_model.name)
    docs += ('   A collection of :py:class:`{0}.{1}` resources for {2}. See'
             '   the'
             '   :py:class:`~boto3.resources.collection.CollectionManager`'
             '   base class for additional methods.\n\n'
             '   This collection uses the :py:meth:`{3}.Client.{4}`'
             '   operation to get items, and its parameters can be'
             '   used as filters::\n\n').format(
                service_name, resource_model.name, official_name,
                service_name, xform_name(collection_model.request.operation))
    docs += ('       for {0} in {1}.{2}.all():\n'
             '           print({0})\n\n').format(
                xform_name(collection_model.resource.type),
                collection_model.parent_name,
                xform_name(collection_model.name),
                xform_name(collection_model.resource.type))

    if collection_model.batch_actions:
        docs += ('   .. rst-class:: admonition-title\n\n   Batch Actions\n\n'
                 '   Batch actions provide a way to manipulate groups of'
                 '   resources in a single service operation call.\n\n')
    for action in sorted(collection_model.batch_actions, key=lambda i:i.name):
        docs += document_action(action, service_name, resource_model,
                                service_model)

    return docs

def document_waiter(waiter, service_name, resource_model, service_model,
                    service_waiter_model):
    """
    Document a resource waiter, including the low-level client waiter
    and parameters.
    """
    try:
        waiter_model = service_waiter_model.get_waiter(waiter.waiter_name)
    except:
        print('Cannot get waiter ' + waiter.waiter_name)
        return ''

    try:
        operation_model = service_model.operation_model(waiter_model.operation)
    except:
        print('Cannot get operation ' + action.request.operation +
              ' for waiter ' + waiter.waiter_name)
        return ''
    description = ('      Waits until this {0} is {1}.\n'
                   '      This method calls ``wait()`` on'
                   ' :py:meth:`{2}.Client.get_waiter` using `{3}`_ .').format(
                        resource_model.name,
                        xform_name(waiter.name).replace('_', ' '),
                        service_name,
                        xform_name(waiter.waiter_name))

    # Here we split because we only care about top-level parameter names
    ignore_params = [p.target.split('.')[0].strip('[]') for p in waiter.params]

    return document_operation(
        operation_model=operation_model, service_name=service_name,
        operation_name=xform_name(waiter.resource_waiter_name),
        description=description,
        example_instance = xform_name(resource_model.name),
        ignore_params=ignore_params, rtype=None)

def document_action(action, service_name, resource_model, service_model,
                    action_type='action'):
    """
    Document a resource action, including the low-level client operation
    and parameters.
    """
    try:
        operation_model = service_model.operation_model(
            action.request.operation)
    except:
        print('Cannot get operation ' + action.request.operation)
        return ''

    # Here we split because we only care about top-level parameter names
    ignore_params = [p.target.split('.')[0].strip('[]')
                     for p in action.request.params]

    rtype = 'dict'
    if action_type == 'action':
        description = ('      This method calls'
                       ' :py:meth:`{0}.Client.{1}`.').format(
                            service_name,
                            xform_name(action.request.operation))
        example_response = 'response'
        if action.resource:
            rtype = ':py:class:`{0}.{1}`'.format(
                service_name, action.resource.type)
            example_response = xform_name(action.resource.type)

            # Is the response plural? If so we are returning a list!
            if action.path and '[]' in action.path:
                rtype = 'list({0})'.format(rtype)

    return document_operation(
        operation_model, service_name, operation_name=xform_name(action.name),
        description=description, ignore_params=ignore_params, rtype=rtype,
        example_instance=xform_name(resource_model.name),
        example_response=example_response)

def document_operation(operation_model, service_name, operation_name=None,
                       description=None, ignore_params=None, rtype='dict',
                       paginated=False, example_instance=None,
                       example_response=None):
    """
    Document an operation. The description can be overridden and certain
    params hidden to support documenting resource actions.
    """
    params = {}
    if operation_model.input_shape:
        try:
            params = operation_model.input_shape.members
        except AttributeError:
            print('Cannot find input shape for ' + operation_model.name)

    if ignore_params is None:
        ignore_params = []

    required = []
    if operation_model.input_shape:
        required = operation_model.input_shape.required_members
    required_params = [k for k in params.keys() if k in required and \
                       k not in ignore_params]
    optional_params = [k for k in params.keys() if k not in required and \
                       k not in ignore_params]
    param_desc = ', '.join([
        ', '.join(['{0}=None'.format(k) for k in required_params]),
        ', '.join(['{0}=None'.format(k) for k in optional_params])
    ])

    if operation_name is None:
        operation_name = xform_name(operation_model.name)

    if description is None:
        description = html_to_rst(
            operation_model._operation_model.get('documentation', ''),
            indent=6, indentFirst=True)

    docs = '   .. py:method:: {0}({1})\n\n{2}\n\n'.format(
                operation_name, param_desc, description)

    if paginated:
        docs += '      This operation can be paginated.\n\n'

    if example_instance:
        dummy_params = []
        for key, value in params.items():
            if key in ignore_params:
                continue
            if key in required_params:
                default = py_default(value.type_name)
                dummy_params.append('{0}={1}'.format(
                    key, default))
        docs +=  '      Example::\n\n          '
        if example_response is not None:
            docs += '{0} = '.format(example_response)
        docs += '{0}.{1}({2})\n\n'.format(example_instance, operation_name,
            ', '.join(dummy_params))

    for key, value in params.items():
        # Skip identifiers as these are automatically set!
        if key in ignore_params:
            continue
        param_type = py_type_name(value.type_name)
        required = key in required_params and 'Required' or 'Optional'
        docs += ('      :param {0} {1}: *{2}* - {3}\n'.format(
            param_type, key, required,
            html_to_rst(value.documentation, indent=9)))
    if rtype is not None:
        docs += '\n\n      :rtype: {0}\n\n'.format(rtype)

    return docs
